//********************************************************************
UNIT dglDataProvider;
// This unit contains:
// 1. TDataProvider: partly abstract object for loading/saving the
//    published properties of TRTTI-Objects (see separate unit).
//    Supported data format: see TRTTIValue
// 2. TCSV_Dataprovider, a concrete descendant of TDataProvider that
//    can load/save TRTTI-Data in CSV-format
// 3. RTTI_ClassList (a TObjectList) that contains TRTTI_Register-
//    records, which are necessary for loading objects. A TRTTI that
//    is to be loaded/saved has to register with RTTI_ClassList.
//
// Missing features:
//  - The content of method pointers (events!)are not loaded/saved
//    as there is no information for finding it (more registration?)
//  - No "LoadChild"-function yet
//  - Errorhandling is not complete
//
// Problems:
// In Delphi: {$M+} is needed for the first ancestor only
// In FPC: every RTTI-Object has to be compiled with {$M+}
//********************************************************************
INTERFACE

USES
   SysUtils, Classes, TypInfo, dglRTTI;
//********************************************************************
TYPE
   TFileAccess = (faLoad, faSave);

   TDGL_DataProvider = Class(TObject)
   Private
      fName: String;            // Filename without extension
      fFileAccess: TFileAccess; // Load/save?
      fOpenSource: Boolean;     // Is DataSource accessable?
      fClient: TRTTI;     // Loaded/saved object

      fClientClassname: String; // internal used variable

      Function CreateLoadedClass: TRTTI;
   Protected
      Function EndOfData: Boolean; Virtual; Abstract;
      Function GetFileName: String; Virtual; Abstract;
      Procedure CloseSource; Virtual; Abstract;
      Procedure OpenSource; Virtual; Abstract;
      Procedure LoadClientHeader; Virtual; Abstract;
      Procedure SaveClientHeader; Virtual; Abstract;
      Procedure LoadProperty(AIndex: Integer); Virtual; Abstract;
      Procedure SaveProperty(AIndex: Integer); Virtual; Abstract;
      Procedure SetDataFormat; Virtual; Abstract;
   Public
      Constructor Create(AName: String;
                   AFileAccess: TFileAccess); Reintroduce;
      Destructor Destroy; Override;

      Function LoadClient: TRTTI;
      Procedure SaveClient(AClient: TRTTI);
   End;

//********************************************************************
   TCSV_DataProvider = Class(TDGL_DataProvider)
   Private
     Lines: TStringList;
     CurLine: Integer;
   Protected
     Function EndOfData: Boolean; Override;
     Function GetFileName: String; Override;
     Procedure CloseSource; Override;
     Procedure OpenSource; Override;
     Procedure LoadClientHeader; Override;
     Procedure SaveClientHeader; Override;
     Procedure LoadProperty(AIndex: Integer); Override;
     Procedure SaveProperty(AIndex: Integer); Override;
     Procedure SetDataFormat; Override;
   End;
//********************************************************************
IMPLEMENTATION
//********************************************************************
// CrackupString cracks a C(omma) S(eparated) V(alue) - string to its
// several parts and fills them in a stringlist
Procedure CrackUpString(AString: String; AStringList: TStringList;
           ASeparator: Char; Const ConsiderEmptyStrings: Boolean);
Var
   FirstPos, LinePos, ItemLen, StrgLen: Word;
   Interim: String;
//--------------------------------------------
Function EOS: Boolean; // End Of String
Begin
   If LinePos <= StrgLen
      Then Result:= False
      Else Result:= True;
End;
//--------------------------------------------
Function EOI: Boolean; // End Of Single Item
Begin
   Result:= True;
   If Not EOS Then
     If AString[LinePos] <> ASeparator
        Then Result:= False
        Else Result:= True;
End;
//--------------------------------------------
Begin
   AStringList.Clear; StrgLen:= Length(AString); LinePos:= 1;
   While Not EOS Do Begin
      FirstPos:= LinePos;
      // scroll to Item-End
      While (Not EOS) And (Not EOI) Do Inc(LinePos);
      // set string, copy and put in list
      ItemLen:= LinePos - FirstPos;
      If (ItemLen > 0) Or (ConsiderEmptyStrings) Then
      Begin
         SetLength(Interim,ItemLen);
         If ItemLen > 0 Then
           Move(AString[FirstPos],Interim[1],ItemLen);
         AStringList.Add(Interim);
      End;
      Inc(LinePos); // = Separator berspringen
   End;
End;
//********************************************************************
// TDGL_DataProvider
//********************************************************************
Constructor TDGL_DataProvider.Create(AName: String;
                               AFileAccess: TFileAccess);
Begin
   Inherited Create;
   fName:= AName;
   fFileAccess:= AFileAccess;
   fClient:= Nil;
   fOpenSource:= False;
   OpenSource;
End;
//********************************************************************
Destructor TDGL_DataProvider.Destroy;
Begin
   CloseSource;
   Inherited Destroy;
End;
//********************************************************************
Function TDGL_DataProvider.CreateLoadedClass: TRTTI;
Var
   LoadedClass: TRTTIClass;
   I: Integer;
Begin
   Result:= Nil;
   LoadedClass:= FindDGLClass(fClientClassName);
   If LoadedClass <> Nil Then
   Begin
      fClient:= LoadedClass.Create;
      SetDataFormat; // String or binary

      fClient.OpenRTTI;
      With fClient Do Begin
         If PropertyCount > 0 Then
           For I:= 0 To PropertyCount-1 Do
             LoadProperty(I);
      End;
      fClient.CloseRTTI;
      Result:= fClient;
   End;
End;
//********************************************************************
Function TDGL_DataProvider.LoadClient: TRTTI;
Begin
   Result:= Nil;

   If fOpenSource Then
   Begin
      LoadClientHeader; // = Load ClassName
      Result:= CreateLoadedClass;
   End;
End;
//********************************************************************
Procedure TDGL_DataProvider.SaveClient(AClient: TRTTI);
var
  I: Integer;
Begin
   fClient:= AClient;

   fClient.OpenRTTI;
   SaveClientHeader; // = Save ClassName
   SetDataFormat;    // String or Binary

   With fClient Do If PropertyCount > 0 Then
     For I:= 0 To PropertyCount-1 Do
       SaveProperty(I);

   fClient.CloseRTTI;
   fClient:= Nil;
End;
//********************************************************************
//********************************************************************
//   TCSV_DataProvider
//********************************************************************
Function TCSV_DataProvider.GetFileName: String;
Begin
   Result:= fName + '.CSV';
End;
//********************************************************************
Procedure TCSV_DataProvider.OpenSource;
Begin
   Lines:= TStringList.Create;

   If fFileAccess = faLoad Then
      If FileExists(GetFileName) Then
      Begin
         Lines.LoadFromFile(GetFileName);
         fOpenSource:= True;
      End;

   CurLine:= 0;
End;
//********************************************************************
Procedure TCSV_DataProvider.CloseSource;
Begin
   If fFileAccess = faSave Then
      Lines.SaveToFile(GetFileName);

   Lines.Free;
End;
//********************************************************************
Function TCSV_DataProvider.EndOfData: Boolean;
Begin
   If (CurLine >= Lines.Count)
      Then Result:= TRUE
      Else Result:= FALSE;
End;
//********************************************************************
Procedure TCSV_DataProvider.LoadClientHeader;
Begin
   fClientClassName:= Lines[CurLine];
   Delete(fClientClassName,Length(fClientClassName),1);
   Inc(CurLine);
End;
//********************************************************************
Procedure TCSV_DataProvider.SaveClientHeader;
Begin
   Lines.Add(fClient.ClassName + ';');
End;
//********************************************************************
Procedure TCSV_DataProvider.LoadProperty(AIndex: Integer);
Var
   CrackList: TStringList;
   OldClient: TRTTI;
   LocalRTTIValue: TRTTIValue;
Begin
   fClient.PropertyIndex:= AIndex; // Load current property
   ClearRTTIValue(LocalRTTIValue);
   LocalRTTIValue.TypeKind:= fClient.CurPropValue.TypeKind;

   CrackList:= TStringList.Create;
   CrackUpString(Lines[CurLine],CrackList,';',True);

   With LocalRTTIValue Do If TypeKind = tkClass
      Then Begin
         OldClient:= fClient;
         VRTTIObject:= LoadClient;
         fClient:= OldClient;
      End
      Else Begin
         VString:= CrackList[1];
         Inc(CurLine);
      End;

   CrackList.Free;

   fClient.CurPropValue:= LocalRTTIValue; // Load property value
End;
//********************************************************************
Procedure TCSV_DataProvider.SaveProperty(AIndex: Integer);
Var OldClient: TRTTI;
Begin
   With fClient Do Begin
      PropertyIndex:= AIndex; // Load current property

   With fClient.CurPropValue Do If TypeKind = tkClass
      Then Begin
         OldClient:= fClient;
         SaveClient(VRTTIObject);
         fClient:= OldClient;
      End
      Else Begin
         Lines.Add(fClient.CurPropName + ';'
                 + fClient.CurPropValue.VString + ';')
      End;
   End;
End;
//********************************************************************
Procedure TCSV_DataProvider.SetDataFormat;
Begin
   fClient.PropDataOutput:= dxString;
End;
//********************************************************************
End.
